/***************************************************************************
 *   Copyright (C) 2006 by Wolfgang Gerlach   *
 *      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/



#include <iostream>
#include <cstdlib>
#include <fstream>
#include <stack>
#include <queue>
#include <functional>
#include <algorithm>
#include <cstring>

#include "dynFMI.h"

using std::cerr;
using std::cout;
using std::endl;
using std::ios;
using std::priority_queue;
using std::vector;



// -------- DynFMI --------

DynFMI::~DynFMI(){
	deleteDynFMINodes(root);
        root = 0;
	delete[] leaves; //free(leaves) // ???;
#if SAMPLE!=0
	delete SampledSAPositionsIndicator;
#endif
//	delete handle;
//	delete pos;
#if SAMPLE!=0
	delete sampledSATree;
#endif
}

void DynFMI::deleteDynFMINodes(WaveletNode *n){
	if (n->right) deleteDynFMINodes(n->right);
	if (n->left) deleteDynFMINodes(n->left);
	delete n;
}

DynFMI::DynFMI(uchar *text, ulong x, ulong n, bool del){
    numberOfTexts = 0;
	initEmptyDynFMI(text, x, n);
	if (del) delete [] text;
}

DynFMI::DynFMI(char *readfile){

	std::ifstream file(readfile);
	if (!file)
	{
		cerr << "error reading file: " << readfile << endl;
		exit(EXIT_FAILURE);
	}
	
	file.seekg(0, ios::end);
	ulong n = file.tellg();
	file.seekg(0, ios::beg);
	
	uchar *text = new uchar[n];
	
	char c;
	ulong i=0;
	
	
	while (file.get(c))
	{
		text[i]=c;
		i++;
	}
	file.close();
	
	initEmptyDynFMI(text, 1, n);
	
	
	delete text;
	
}

void DynFMI::iterateReset(){
	iterate = 1;
	recursiveIterateResetWaveletNode(root);
}

void DynFMI::recursiveIterateResetWaveletNode(WaveletNode *w){
	w->bittree->iterateReset();
	
	if (w->left) recursiveIterateResetWaveletNode(w->left);
	if (w->right) recursiveIterateResetWaveletNode(w->right);
}


bool DynFMI::iterateNext(){
	iterate++;
	return !(iterate > getSize());
}

uchar DynFMI::iterateGetSymbol(){

	ulong i = iterate;
	WaveletNode *walk = root;	
	bool bit;
		
	while (true) {
		
		bit = walk->bittree->iterateGetBit(); // TODO improve
		i=walk->bittree->iterateGetRank(bit);
		
		walk->bittree->iterateNext();
		
		
		if (bit) { //bit = 1
			if (walk->right == 0) return walk->c1;
			walk=walk->right;
		} else { // bit = 0
			if (walk->left == 0) return walk->c0;
			walk=walk->left;
		}
		
		
	} // end of while
	
}


uchar* DynFMI::getBWT(){
	ulong n = root->bittree->getPositions();
	n++;
	uchar *text = new uchar[n];
	
	bool data=true;
	// old slow version:
	//for (ulong i=1; i <= root->bittree->getPositions(); i++)
	//	text[i-1]=(*this)[i];
	
	ulong i = 0;
	
	iterateReset();
	
	while (data) {
		text[i] = iterateGetSymbol();	
		data = iterateNext();
		i++;
	}
	text[n-1]=0;
	
	return text;
}

std::ostream& DynFMI::getBWTStream(std::ostream& stream){
	ulong n = root->bittree->getPositions();
	n++;
	//uchar *text = new uchar[n];
	
	bool data=true;
	// old slow version:
	//for (ulong i=1; i <= root->bittree->getPositions(); i++)
	//	text[i-1]=(*this)[i];
	
	ulong i = 0;
	
	iterateReset();
	
	while (data) {
		//text[i] = iterateGetSymbol();	
		stream << iterateGetSymbol();
		data = iterateNext();
		i++;
	}
	//text[n-1]=0;

	return stream;
}

//void DynFMI::printDynFMIContent(ostream& stream){
//	uchar c;
//	for (ulong i=1; i<=getSize(); i++) 
//	{
//		c =(*this)[i];
//		if (c==0) c= '#';
//		stream << c;
//	}


void DynFMI::deleteLeaves(WaveletNode *node){
	bool leaf = true;

	if (node->left) {
		// internal node
		leaf = false;
		deleteLeaves(node->left);
		
	}
	if (node->right){
		leaf = false;
		deleteLeaves(node->right);
	} 
	
	if (leaf) {
		// is a leaf, delete it!
		if (node->parent) {
			if (node==node->parent->left) node->parent->left=0;
				else node->parent->right=0;
		}
		delete node;
	}
}

void DynFMI::makeCodes(ulong code, int bits, WaveletNode *node){
	#ifndef NDEBUG
	if (node == node->left) {
		cout << "makeCodes: autsch" << endl;
		exit(0);
		}
	#endif

	if (node->left) {
		makeCodes(code | (0 << bits), bits+1, node->left);
		makeCodes(code | (1 << bits), bits+1, node->right);
	} else {
		codes[node->c0] = code;
		codelengths[node->c0] = bits+1;
	}
}

void DynFMI::appendBVTrees(WaveletNode *node){
	node->bittree = new BVTree();

	if (node->left) appendBVTrees(node->left);
	if (node->right) appendBVTrees(node->right);
}

void DynFMI::initEmptyDynFMI(uchar *text, ulong x, ulong n){
	// pointers to the leaves for select
	leaves = (WaveletNode**) new WaveletNode*[256];
	for(int j=0; j<256; j++) leaves[j]=0;


	ulong i;
	for(i=0; i < n; i++) {
		if (text[i]==0) {
			cerr << "Error: unsigned char 0 is reserved." << endl;
			exit(EXIT_FAILURE);
		}
		
		if (leaves[text[i]]==0) {
			leaves[text[i]] = new WaveletNode(text[i]); 
		}
		leaves[text[i]]->weight++; 
	}
	
	// separation symbol:
	leaves[0] = new WaveletNode((uchar)0); 
	leaves[0]->weight=x;
	
	// Veli's approach:
	priority_queue< WaveletNode*, vector<WaveletNode*>, std::greater<WaveletNode*> > q;
	
	
	for(int j=255; j>=0; j--){
		if (leaves[j]!=0) {
			q.push( (leaves[j]) );
		}
		codes[j] = 0;
		codelengths[j] = 0;
	}
	
	// creates huffman shape:
	while (q.size() > 1) {
		
		WaveletNode *left = q.top();
		q.pop();
		
		WaveletNode *right = q.top();
		q.pop();
		
		q.push(  new WaveletNode(left, right) );
	}	
	

	root = q.top();
	q.pop();
	
			
	makeCodes(0,0, root);	// writes codes and codelengths

	
	// merge leaves	(one leaf represent two characters!)
	for(int j=0; j<256; j++){
	
		if (leaves[j]) {
		
			if (leaves[j]->parent->left==leaves[j]) {
				leaves[j]->parent->c0=j;
			} else {
				leaves[j]->parent->c1=j;
			}
			leaves[j]=leaves[j]->parent; // merge
		}
	}
	
	deleteLeaves(root);
	appendBVTrees(root);
	
	// array C needed for backwards search
	for(int j=0; j<256+256; j++) C[j] = 0;
	
#if SAMPLE!=0
	this->SampledSAPositionsIndicator = new BVTree();
	this->sampledSATree = new SampledSATree();
#endif
	
	//this->pos = new Pos(sampleInterval);
	//this->handle = new Handle();
	
	//pos->handle=this->handle;
	//handle->pos=this->pos;
	
}

void DynFMI::append(uchar *text, ulong length){
	ulong i;
	ulong start = root->bittree->getPositions();
	for (i = start; i<start+length; i++)
		insert(text[i-start], i+1);
}


void DynFMI::append(uchar c){
	insert(c, root->bittree->getPositions()+1);
}

void DynFMI::insert(uchar c, ulong i){
	#ifndef NDEBUG
	if (codelengths[c]==0) {
		cerr << "error: Symbol \"" << c << "\" (" << (int)c << ") is not in the code table!" << endl;;
		exit(EXIT_FAILURE);
	}
	#endif
	
	ulong level = 0;
	ulong code = codes[c];

	bool bit;
	WaveletNode *walk = root;	
		
	while (walk) {
		
		bit = ((code & (1u << level)) != 0); 
		
		walk->bittree->insertBit(bit,i); // TODO improve
		i=walk->bittree->rank(bit, i);

		if (bit) { //bit = 1
			walk=walk->right;
		} else { // bit = 0
			walk=walk->left;
		}
		
		level++;		
	} // end of while
	
	int j = 256+c;
	while(j>1) {
		C[j]++;
		j=binaryTree_parent(j);
		}
	C[j]++;	
	
}

void DynFMI::deleteSymbol(ulong i){
	WaveletNode *walk = root;	
	bool bit;
	uchar c;
	while (true) {

		// original slow version:
		//bit = (*walk->bittree)[i];
		//old_i = i;		
		//i=walk->bittree->rank(bit, i);	
		//walk->bittree->deleteBit(old_i);


		walk->bittree->deleteBit(i);
		bit=walk->bittree->getLastBitDeleted();
		i=walk->bittree->getLastRank();
		
		if (bit) { //bit = 1
			if (walk->right == 0) {
				c = walk->c1;
				break;
				}
			walk=walk->right;
		} else { // bit = 0
			if (walk->left == 0) {
				c = walk->c0;
				break;
			}
			walk=walk->left;
		}
		
		
	} // end of while
	
	int j = 256+c;
	while(j>1) {
		C[j]--;
		j=binaryTree_parent(j);
		}
	C[j]--;	
	
	
}


uchar DynFMI::operator[](ulong i){
	WaveletNode *walk = root;	
	bool bit;
		
	while (true) {
		
		bit = (*walk->bittree)[i]; //TODO improve by reducing
		i=walk->bittree->rank(bit, i);

		if (bit) { //bit = 1
			if (walk->right == 0) return walk->c1;
			walk=walk->right;
		} else { // bit = 0
			if (walk->left == 0) return walk->c0;
			walk=walk->left;
		}
		
		
	} // end of while
	cout << endl;
}

ulong DynFMI::rank(uchar c, ulong i){
	if (i==0) return 0;

	ulong level = 0;
	ulong code = codes[c];
	
	
	bool bit;
	WaveletNode *walk = root;	
		
	while (true) {
		
		bit = ((code & (1u << level)) != 0);
		
		i=walk->bittree->rank(bit, i);
		if (bit) { //bit = 1
			if (walk->right == 0) return i;
			walk=walk->right;
		} else { // bit = 0
			if (walk->left == 0) return i;
			walk=walk->left;
		}
	
		level++;		
	} // end of while
	
	cerr << "error: DynFMI::rank: left while loop" << endl;
	exit(EXIT_FAILURE);
	return 0; //never
}

ulong DynFMI::select(uchar c, ulong i){
	
	WaveletNode *walk = leaves[c];	
	
	bool bit = (walk->c1==c);
	
	while (walk->parent) {
		i=walk->bittree->select(bit, i);
		
		bit = (walk == walk->parent->right);
		walk=walk->parent;
	} // end of while
	
	i=walk->bittree->select(bit, i);

	return i;
}


ulong DynFMI::count(uchar *pattern){

	ulong i = strlen((char*)pattern);
	ulong sp = 1;

	ulong ep = getSize();
	uchar s;
	while ((sp <= ep) && (i >=1)) {
		s=pattern[i-1];
		sp=(ulong)getNumberOfSymbolsSmallerThan(s) + rank(s, sp-1) + 1;
		ep=(ulong)getNumberOfSymbolsSmallerThan(s) + rank(s,ep);
		
		i--;
	}
	return ep-sp +1;
}

#if SAMPLE!=0
ulong* DynFMI::locate(uchar *pattern){

	ulong i = strlen((char*)pattern);
	ulong sp = 1;

	ulong ep = getSize();
	uchar s;
	while ((sp <= ep) && (i >=1)) {
		s=pattern[i-1];
		sp=(ulong)getNumberOfSymbolsSmallerThan(s) + rank(s, sp-1) + 1;
		ep=(ulong)getNumberOfSymbolsSmallerThan(s) + rank(s,ep);
		
		i--;
	}

	ulong numberOfMatches = ep-sp +1;
	ulong *matches = new ulong[2*numberOfMatches + 1];
	matches[0] = numberOfMatches;

	ulong diff;
	ulong k = 1;
	cout << "----------------------" << endl;	
	for (ulong j=sp; j <= ep; j++) {

		i=j;
		diff = 0;
		while (!(*SampledSAPositionsIndicator)[i]) {
			diff++;
			s=(*this)[i];
			i= (ulong)getNumberOfSymbolsSmallerThan(s) + rank(s, i);
	
		}
	

		sampledSATree->getSample(SampledSAPositionsIndicator->rank(true, i));

		//cout << "found: (" << sampledSATree->handle  << ", " << sampledSATree->sampledSAValue + diff << ")"<< endl;
		matches[k++] = sampledSATree->handle;
		matches[k++] = sampledSATree->sampledSAValue + diff;
	}

	return matches;
}
#endif

// size must include endmarker!
void DynFMI::addText(uchar const * str, ulong n){
	ulong i;
#if SAMPLE!=0
	bool sample = false;
#endif
	i = numberOfTexts + 1;
 
	numberOfTexts ++;

        // Adding an empty text (end-marker)
        if (n == 1)
        {
            insert(str[0], i);
            //if (del) delete [] str;
            return;
        }

	insert(str[n-2],i); // insert second last character, corresponds to suffix of length 1
#if SAMPLE!=0
	sample = (n-1 % sampleInterval == 0);
	SampledSAPositionsIndicator->insertBit(sample, i);	
	if (sample) sampledSATree->insertSample(n-1, key, SampledSAPositionsIndicator->rank(true,i));
#endif
	for (ulong t=n-2; t > 0; t--) {
		i= 1+getNumberOfSymbolsSmallerThan(str[t]) + rank(str[t],i);
		insert(str[t-1],i);
#if SAMPLE!=0
		sample = ((t) % sampleInterval == 0);
		SampledSAPositionsIndicator->insertBit(sample, i);

		if (sample) sampledSATree->insertSample(t, key, SampledSAPositionsIndicator->rank(true,i));
#endif
	}
	

	i= 1+ getNumberOfSymbolsSmallerThan(str[0]) + rank(str[0],i);
	insert(str[n-1],i);
#if SAMPLE!=0
	SampledSAPositionsIndicator->insertBit(true, i); 
	sampledSATree->insertSample(1, key, SampledSAPositionsIndicator->rank(true,i));
#endif	
//	if (del)  delete [] str;
	
//	return key;
}


/*ulong DynFMI::addTextFromFile(char* filename, ulong n){
	ulong i;
#if SAMPLE!=0
	bool sample = false;
#endif
	i=pos->getSize()+1;


	std::ifstream str(filename, ios::binary);
	if (!str)
	{
		cerr << "error reading file: " << filename << endl;
		exit(EXIT_FAILURE);
	}

	char c;
	//char oldc;


	ulong buf_len; //1048576; //10MB
	char * buffer;
	ulong buf_pos;

	if (BUFFER==0) buf_len = n;
		else buf_len = BUFFER;

	buffer = new char[buf_len];
	buf_pos = (n-2) / buf_len;
	str.seekg(buf_pos*buf_len);
	str.read(buffer, buf_len);
	str.clear(std::ios_base::goodbit);


	ulong key = pos->appendText(n);
	
	//str.seekg(n-2);

	//c=str.get();
	c=buffer[(n-2)%buf_len];
	insert(c,i); // insert second last character, corresponds to suffix of length 1
#if SAMPLE!=0
	sample = (n-1 % sampleInterval == 0);
	SampledSAPositionsIndicator->insertBit(sample, i);	
	if (sample) sampledSATree->insertSample(n-1, key, SampledSAPositionsIndicator->rank(true,i));
#endif
	if ((n-2)%buf_len == 0) {
		c=buffer[0];
		buf_pos--;
		str.seekg(buf_pos*buf_len);
		str.read(buffer, buf_len);
	}
	for (ulong t=n-2; t > 0; t--) {

		i= 1+getNumberOfSymbolsSmallerThan(c) + rank(c,i);

		c=buffer[(t-1)%buf_len];
		insert(c,i);

		//check buffer
		if (((t-1)%buf_len == 0) && ((t-1)!=0)) {
#ifndef NDEBUG
			if (buf_pos == 0) {
				cerr << "buf_pos too small" << endl;
				exit(0);
			}
#endif
			buf_pos--;
			str.seekg(buf_pos*buf_len);
			str.read(buffer, buf_len);
		}

#if SAMPLE!=0
		sample = ((t) % sampleInterval == 0);
		SampledSAPositionsIndicator->insertBit(sample, i);

		if (sample) sampledSATree->insertSample(t, key, SampledSAPositionsIndicator->rank(true,i));
#endif
	}

	i= 1+ getNumberOfSymbolsSmallerThan(c) + rank(c,i);
	insert(c,i);

#if SAMPLE!=0
	SampledSAPositionsIndicator->insertBit(true, i); 
	sampledSATree->insertSample(1, key, SampledSAPositionsIndicator->rank(true,i));
#endif	

	str.close();

	return key;
}
*/
/*
uchar* DynFMI::retrieveText(ulong key){
	cout << "key: " << key << endl;
	uchar *text=0;
	// TODO access two times, too much
	ulong i=handle->getPos(key); //=bwtEnd;
	if (i == 0 ) {
		return text;
	}

	ulong n=pos->getTextSize(i);
	
	text = new uchar[n]; // last byte 0 for cout
	
	uchar c;
	
//TODO better	
	for (ulong t=n-2; t < n; t--) {
		c = (*this)[i];
		text[t]=(c==0?'#':c);
		
		i= 0+ getNumberOfSymbolsSmallerThan(c) + rank(c,i);
	}

	#ifndef NDEBUG		
	for (ulong i=0; i< n-1 ; i++) if (text[i]==0) text[i]='-';
	#endif	
	
	text[n-1] = 0;
	return text;
        }*/

/*
bool DynFMI::deleteText(ulong key){
	// TODO access two times, can do better
	ulong i=handle->getPos(key); //=bwtEnd;
	if (i == 0) return false;

	ulong n=pos->getTextSize(i);
	ulong *text = new ulong[n]; 
	uchar c;
	
	for (ulong t=n-1; t <n; t--) {
		c = (*this)[i];
		text[t]=i;
		i= 0+ getNumberOfSymbolsSmallerThan(c) + rank(c,i); // TODO improve by lastrank!
	}


        std::sort(text, text + n);


	for (i=n-1; i<n; i--) {
		this->deleteSymbol(text[i]); 
#if SAMPLE!=0
		SampledSAPositionsIndicator->deleteBit(text[i]);		
//TODO samples ?
#endif
	}
	delete[] text;
	handle->deleteKey(key);
	
		
	return true;
}
*/
ulong DynFMI::getNumberOfSymbolsSmallerThan(uchar c){
	int j = 256+c;
	ulong r=0;
	while(j>1) {
		if (binaryTree_isRightChild(j)) 
			r += C[binaryTree_left(binaryTree_parent(j))];
		
		j=binaryTree_parent(j);
	}
	return r;
}




void DynFMI::printDynFMIContent(std::ostream& stream){
	uchar c;
	for (ulong i=1; i<=getSize(); i++) 
	{
		c =(*this)[i];
		if (c==0) c= '#';
		stream << c;
	}
}




