#include <cstdlib>
extern "C" {
#include <unistd.h>
}
#include <bp/bp.h>
#include "common_stub.hpp"

extern "C" value caml_bitmap_create(value size)
{
  CAMLparam1(size);
  size_t bits = Long_val(size);
  size_t words = bits / (8*sizeof(unsigned int));
  unsigned int *buffer = (unsigned int*) calloc(words+1, sizeof(unsigned int));
  if (buffer == NULL)
    sxsi_raise_msg("BP: cannot allocate memory");
  CAMLreturn( (value) buffer);
}

extern "C" value caml_bitmap_resize(value bitmap, value nsize)
{
  CAMLparam2(bitmap, nsize);
  size_t bits = Long_val(nsize);
  size_t bytes = (bits / (8 * sizeof(unsigned int)) + 1 ) * sizeof(unsigned int);
  unsigned int * buffer = (unsigned int*) realloc((void *) bitmap, bytes);
  if (buffer == NULL)
    sxsi_raise_msg("BP: cannot reallocate memory");
  CAMLreturn((value) buffer);
}

extern "C" value caml_bitmap_setbit(value bitmap, value i, value b)
{
  CAMLparam3(bitmap, i, b);
  unsigned int j = Int_val(i);
  unsigned int x = Bool_val(b);
  bp_setbit ((unsigned int*) bitmap, j, x);
  CAMLreturn(Val_unit);
}

extern "C" void caml_bp_delete(value b)
{
  CAMLparam1(b);
  bp * B = Obj_val<bp*>(b);
  bp_delete(B);
  CAMLreturn0;
}

extern "C" value caml_bp_construct(value bitmap, value npar)
{
  CAMLparam2(bitmap, npar);
  CAMLlocal1(res);
  bp * b = bp_construct(Int_val(npar), (unsigned int *) bitmap, OPT_DEGREE);
  res = sxsi_alloc_custom<bp*>(caml_bp_delete);
  Obj_val<bp*>(res) = b;
  CAMLreturn(res);
}

extern "C" value caml_bp_first_child(value b, value idx)
{
  CAMLparam2(b, idx);
  CAMLreturn (Val_int( bp_first_child(Obj_val<bp*>(b), Int_val(idx))));
}


extern "C" value caml_bp_next_sibling(value b, value idx)
{
  CAMLparam2(b, idx);
  CAMLreturn (Val_int(bp_next_sibling(Obj_val<bp*>(b), Int_val(idx))));
}

extern "C" value caml_bp_preorder_rank(value b, value idx)
{
  CAMLparam2(b, idx);
  CAMLreturn (Val_int(bp_preorder_rank(Obj_val<bp*>(b), Int_val(idx)) - 1));
}


extern "C" value caml_bp_load(value file)
{
  CAMLparam1(file);
  CAMLlocal1(result);
  bp *B;
  int f1 = Int_val(file);
  int f2 = dup(f1);
  FILE * fd = fdopen(f2, "r");
  if (fd == NULL)
    sxsi_raise_msg("Error opening bp file");
  B = loadTree(fd);
  fclose(fd);
  result = sxsi_alloc_custom<bp*>(caml_bp_delete);
  Obj_val<bp*>(result) = B;
  CAMLreturn(result);
}

extern "C" value caml_bp_save(value b, value file)
{
  CAMLparam2(b, file);
  bp *B = Obj_val<bp*>(b);
  int f1 = Int_val(file);
  int f2 = dup(f1);
  FILE * fd = fdopen(f2, "a");
  fflush(stderr);
  if (fd == NULL)
    sxsi_raise_msg("Error saving bp file");
  saveTree(B, fd);
  fclose(fd);
  CAMLreturn(Val_unit);
}

extern "C" value caml_bp_alloc_stats(value unit)
{
  CAMLparam1(unit);
  CAMLreturn (Val_long(bp_get_alloc_stats()));
}
