#include "xml-tree.hpp"
#include "utils_stub.hpp"
#include <cstdio>

using namespace SXSI;

static xml_tree*& XMLTREE(value v)
{
  return Obj_val<xml_tree*>(v);
}

static xml_tree::node_t TREENODE(value i)
{
  return static_cast<xml_tree::node_t>(Int_val(i));
}

static xml_tree::tag_t TAG(value i)
{
  return static_cast<xml_tree::tag_t>(Int_val(i));
}

extern "C"  value caml_xml_tree_save(value tree, value fd, value prefix)
{
  CAMLparam3(tree, fd, prefix);
  XMLTREE(tree)->save(Int_val(fd), String_val(prefix));
  CAMLreturn (Val_unit);
}

extern "C" value
caml_xml_tree_load(value fd, value prefix, value load_tc, value sf)
{
  CAMLparam4(fd, prefix, load_tc, sf);
  CAMLlocal1(result);
  xml_tree * tree;
  try {

    tree = xml_tree::load(Int_val(fd),
                          String_val(prefix),
                          Bool_val(load_tc),
                          Int_val(sf));

    result = sxsi_alloc_custom<xml_tree*>();
    XMLTREE(result) = tree;
    CAMLreturn(result);
  }
  catch (const std::exception& e){ sxsi_raise_msg(e.what()); }
  catch (std::string msg){ sxsi_raise_msg(msg.c_str()); }
  catch (char const * msg){ sxsi_raise_msg(msg);  };
  //never reached
  return (Val_unit);
}

NoAlloc extern "C"  value caml_xml_tree_root(value tree)
{
  return (Val_int(XMLTREE(tree)->root()));
}

NoAlloc extern "C"  value caml_xml_tree_size(value tree)
{
  return (Val_int(XMLTREE(tree)->size()));
}

NoAlloc extern "C"  value caml_xml_tree_num_tags(value tree)
{
  return (Val_int(XMLTREE(tree)->num_tags()));
}

NoAlloc extern "C"  value caml_xml_tree_subtree_size(value tree, value node)
{
  return (Val_int(XMLTREE(tree)->subtree_size(TREENODE(node))));
}

NoAlloc extern "C" value
caml_xml_tree_subtree_tags(value tree, value node, value tag)
{
  return (Val_int(XMLTREE(tree)->subtree_tags(TREENODE(node),
                                              TAG(tag))));
}

NoAlloc extern "C" value caml_xml_tree_subtree_elements(value tree, value node,
							value atts)
{
  return (Val_int(XMLTREE(tree)->subtree_elements(TREENODE(node),
						  TAGLIST(atts))));
}

NoAlloc extern "C" value caml_xml_tree_is_leaf(value tree, value node){
  return (Val_bool(XMLTREE(tree)->is_leaf(TREENODE(node))));
}

NoAlloc extern "C" value
caml_xml_tree_is_ancestor(value tree, value node1, value node2)
{
  return (Val_bool(XMLTREE(tree)->is_ancestor(TREENODE(node1),
                                              TREENODE(node2))));
}

NoAlloc extern "C" value
caml_xml_tree_is_child(value tree, value node1, value node2)
{
  return (Val_bool(XMLTREE(tree)->is_child(TREENODE(node1),
                                           TREENODE(node2))));
}

NoAlloc extern "C" value caml_xml_tree_is_first_child(value tree, value node)
{
  return (Val_bool(XMLTREE(tree)->is_first_child(TREENODE(node))));
}

NoAlloc extern "C" value
caml_xml_tree_is_right_descendant(value tree, value x, value y)
{
  return (Val_bool(XMLTREE(tree)->is_right_descendant(TREENODE(x),
                                                      TREENODE(y))));
}

NoAlloc extern "C"  value caml_xml_tree_num_children(value tree, value node)
{
  return (Val_int(XMLTREE(tree)->num_children(TREENODE(node))));
}

NoAlloc extern "C"  value caml_xml_tree_child_pos(value tree, value node)
{
  return (Val_int(XMLTREE(tree)->child_pos(TREENODE(node))));
}

NoAlloc extern "C"  value caml_xml_tree_depth(value tree, value node)
{
  return (Val_int(XMLTREE(tree)->depth(TREENODE(node))));
}

NoAlloc extern "C"  value caml_xml_tree_preorder(value tree, value node)
{
  return (Val_int(XMLTREE(tree)->preorder(TREENODE(node))));
}

NoAlloc extern "C"  value caml_xml_tree_postorder(value tree, value node)
{
  return (Val_int(XMLTREE(tree)->postorder(TREENODE(node))));
}

NoAlloc extern "C"  value caml_xml_tree_tag(value tree, value node)
{
  return (Val_int(XMLTREE(tree)->tag(TREENODE(node))));
}

NoAlloc extern "C"  value caml_xml_tree_parent(value tree, value node)
{
  return (Val_int(XMLTREE(tree)->parent(TREENODE(node))));
}

NoAlloc extern "C"  value caml_xml_tree_child(value tree, value node, value idx)
{
  return (Val_int(XMLTREE(tree)->child(TREENODE(node), Int_val(idx))));
}

NoAlloc extern "C"  value caml_xml_tree_first_child(value tree, value node)
{
  return (Val_int(XMLTREE(tree)->first_child(TREENODE(node))));
}

NoAlloc extern "C"  value caml_xml_tree_first_element(value tree, value node)
{
  return (Val_int(XMLTREE(tree)->first_element(TREENODE(node))));
}

NoAlloc extern "C"  value caml_xml_tree_last_child(value tree, value node)
{
  return (Val_int(XMLTREE(tree)->last_child(TREENODE(node))));
}

NoAlloc extern "C"  value caml_xml_tree_next_sibling(value tree, value node)
{
  return (Val_int(XMLTREE(tree)->next_sibling(TREENODE(node))));
}

NoAlloc extern "C" value caml_xml_tree_next_element(value tree, value node)
{
  return (Val_int(XMLTREE(tree)->next_element(TREENODE(node))));
}

NoAlloc extern "C" value caml_xml_tree_prev_sibling(value tree, value node)
{
  return (Val_int(XMLTREE(tree)->prev_sibling(TREENODE(node))));
}

NoAlloc extern "C" value
caml_xml_tree_tagged_child(value tree, value node, value tag)
{
  return (Val_int(XMLTREE(tree)->tagged_child(TREENODE(node),
                                              TAG(tag))));
}

NoAlloc extern "C" value
caml_xml_tree_select_child(value tree, value node, value tags)
{
  return (Val_int(XMLTREE(tree)->select_child(TREENODE(node), TAGLIST(tags))));
}

NoAlloc extern "C" value
caml_xml_tree_tagged_sibling(value tree, value node, value tag)
{
  return (Val_int(XMLTREE(tree)->tagged_sibling(TREENODE(node),
                                                TAG(tag))));
}

NoAlloc extern "C" value
caml_xml_tree_select_sibling(value tree, value node, value tags)
{
  return (Val_int(XMLTREE(tree)->select_sibling(TREENODE(node),
                                                TAGLIST(tags))));
}

NoAlloc extern "C" value
caml_xml_tree_tagged_descendant(value tree, value node, value tag)
{
  return (Val_int(XMLTREE(tree)->tagged_descendant(TREENODE(node),
                                                   TAG(tag))));
}

NoAlloc extern "C" value
caml_xml_tree_tagged_next(value tree, value node, value tag)
{
  return (Val_int(XMLTREE(tree)->tagged_next(TREENODE(node),
                                             TAG(tag))));
}

NoAlloc extern "C" value
caml_xml_tree_select_descendant(value tree, value node, value tags)
{
  return (Val_int(XMLTREE(tree)->select_descendant(TREENODE(node),
                                                   TAGLIST(tags))));
}

NoAlloc extern "C" value caml_xml_tree_tagged_following_before(value tree,
                                                               value node,
                                                               value tag,
                                                               value closing)
{
  return (Val_int(XMLTREE(tree)->tagged_following_before(TREENODE(node),
                                                         TAG(tag),
                                                         TREENODE(closing))));
}

NoAlloc extern "C"  value caml_xml_tree_select_following_before(value tree,
                                                                value node,
                                                                value tags,
                                                                value closing)
{
  return (Val_int(XMLTREE(tree)->select_following_before(TREENODE(node),
                                                         TAGLIST(tags),
                                                         TREENODE(closing))));
}



extern "C" value caml_xml_tree_get_text_collection(value tree)
{
  CAMLparam1(tree);
  CAMLlocal1(text);
  text = sxsi_alloc_custom<TextCollection*>();
  Obj_val<TextCollection*>(text) = XMLTREE(tree)->get_text_collection();
  CAMLreturn (text);
}

NoAlloc extern "C" value caml_xml_tree_closing(value tree, value node)
{
  return (Val_int(XMLTREE(tree)->closing(TREENODE(node))));
}

NoAlloc extern "C"  value caml_xml_tree_nullt(value unit){
  return (Val_int(xml_tree::NIL));
}


extern "C" value caml_xml_tree_print(value tree, value node, value fd)
{
  CAMLparam3(tree, node, fd);
  XMLTREE(tree)->print(TREENODE(node), Int_val(fd));
  CAMLreturn(Val_unit);
}


extern "C" value caml_xml_tree_get_tag_name(value tree, value tag)
{
  CAMLparam2(tree, tag);
  CAMLlocal1(res);
  const char* s = XMLTREE(tree)->get_tag_name_by_ref(TAG(tag));
  res = caml_copy_string(s);
  CAMLreturn(res);
}

NoAlloc extern "C" value caml_xml_tree_flush(value tree, value fd)
{
  XMLTREE(tree)->flush(Int_val(fd));
  return Val_unit;
}

extern "C" value caml_xml_tree_register_tag(value tree, value str)
{
  CAMLparam2(tree, str);
  value res;
  res = Val_int(XMLTREE(tree)->register_tag(String_val(str)));
  CAMLreturn(res);
}


/** Full reporting into a bit vector
 */
static void pr_vector(std::vector<int32_t> v)
{
  for(auto i = v.begin(); i!= v.end(); ++i)
    fprintf(stderr, "%i ", *i);
  fprintf(stderr, "\n");
}
static std::vector<int32_t> sort_results(std::vector<int32_t> v, xml_tree *t)
{
  std::vector<int32_t> res;
  std::sort(v.begin(), v.end());
  int32_t prev = -1;
  for(auto i = v.begin(); i != v.end(); ++i){
    while (prev == *i){
      ++i;
      if (i == v.end()) return res;
    };
    prev = *i;
    res.push_back(t->parent_node(prev));
  };
  return res;
}

#define BV_QUERY(pref, Pref) \
  extern "C" value caml_text_collection_## pref ##_bv(value tree, value str, value dobvv){ \
    CAMLparam3(tree, str, dobvv);					\
    CAMLlocal3(res, res_bv, res_array);					\
    int j;								\
    uchar * cstr = (uchar *) strdup(String_val(str));			\
    std::vector<int32_t> uresults = XMLTREE(tree)->Pref(cstr);		\
    std::vector<int32_t> results = sort_results(uresults, XMLTREE(tree)); \
    bool dobv = Bool_val(dobvv);					\
    res_bv = caml_alloc_string(dobv ? ((XMLTREE(tree)->size() / 4) + 2) : 0); \
    unsigned long slen = caml_string_length(res_bv);			\
    if (dobv)								\
      memset(&(Byte(res_bv,0)), 0, slen);				\
    res_array = caml_alloc_shr(results.size(), 0);			\
    for (unsigned int i = 0; i < results.size(); ++i) {			\
      j = results[i];							\
      if (dobv)	{							\
	Byte(res_bv, j >> 3) |=   (1 << (j & 7));			\
      };								\
      caml_initialize(&Field(res_array, i), Val_int(j));		\
    };									\
    free(cstr);								\
    res = caml_alloc(2, 0);						\
    Store_field(res, 0, res_bv);					\
    Store_field(res, 1, res_array);					\
    CAMLreturn(res);							\
  }									\


BV_QUERY(prefix, prefix)
BV_QUERY(suffix, suffix)
BV_QUERY(equals, equals)
BV_QUERY(contains, contains)
BV_QUERY(lessthan, less_than)


extern "C" value caml_xml_tree_naive_contains(value tree, value node, value word)
{
  CAMLparam3(tree, node, word);
  const unsigned char * s = (const unsigned char*) String_val(word);
  CAMLreturn (Val_bool(XMLTREE(tree)->naive_contains(TREENODE(node), s)));
}
